/* 
   File: frame.cpp - Implementation of main frame window.

   Implements ZDDesk's main frame window.  This window is never visible it is
   mainly provided to handle tray notification messages and to display the
   context menu and to process menu commands.  It implements the "virtual 
   desk" code for ZDDesk -- i.e. the switching of desktops, etc..

   Author: Steven E. Sipe
   
*/

#include "stdafx.h"
#include "zddesk.h"

#include "frame.h"
#include "splash.h"

#include "tray.h"
#include "registry.h"
#include "arrange.h"
#include "property.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

static const char *ZDDESKKEY = "Software\\PC Magazine\\ZDDesk";
static const int UM_TRAYNOTIFY = WM_USER+2000;

/////////////////////////////////////////////////////////////////////////////
// CZDDeskFrame

BEGIN_MESSAGE_MAP(CZDDeskFrame, CFrameWnd)
   //{{AFX_MSG_MAP(CZDDeskFrame)
   ON_WM_CREATE()
   ON_WM_DESTROY()
   ON_WM_RBUTTONUP()
   ON_COMMAND(IDM_EXIT, OnExit)
   ON_WM_LBUTTONUP()
   ON_COMMAND(IDM_ARRANGE, OnArrange)
   ON_COMMAND(IDM_PROPERTIES, OnProperties)
   ON_COMMAND(IDM_ABOUT, OnAbout)
   ON_COMMAND(IDM_HELP, OnHelp)
   //}}AFX_MSG_MAP
   ON_MESSAGE(UM_TRAYNOTIFY,OnTrayNotify)
   ON_COMMAND_RANGE(IDM_DESK1,IDM_DESK9,OnSelectDesk)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CZDDeskFrame construction/destruction

// Constructor
CZDDeskFrame::CZDDeskFrame()
{
   // Initialize the virtual desktop
   InitializeDesks();
}

// Destructor
CZDDeskFrame::~CZDDeskFrame()
{
   // Shutdown the virutual desktop
   UninitializeDesks();
}

///////////////////////////////
/// "Safe" utility routines ///
///////////////////////////////

// Performs a ShowWindow() call without locking up on frozen applications.
// Basically, the WM_SHOWWINDOW message generated for each app by ShowWindow() 
// locks the sending app under Windows if the receiving app is dead
BOOL CZDDeskFrame::SafeShowWindow(HWND hWnd, int nShowCmd)
{
   // Determine if the app is still alive or hung
   DWORD dwResult;
   if(::SendMessageTimeout(hWnd,WM_NULL,0,0,SMTO_ABORTIFHUNG,1000,&dwResult))
   {
      // Seemed to work okay --- show the window
      if(dwResult == 0)
         ::ShowWindow(hWnd,nShowCmd);
      return(TRUE);
   }
   else 
   {
      // Skating on thin ice now, don't send anymore color changes
      if(!m_bFrozen)
      {
         // Warn the user about the potential problem
         m_bFrozen = TRUE;
         AfxMessageBox(IDS_FROZEN,MB_ICONWARNING|MB_OK);
      }
      return(FALSE);
   }
}

// Don't set the system colors if we have detected a dead application --
// the WM_SYSCOLORCHANGE message also locks up with dead applications
void CZDDeskFrame::SafeSetSysColors(int cElements, CONST INT *lpaElements,
                              CONST COLORREF *lpaRgbValues)
{
   // Okay to send???
   if(!m_bFrozen)
      ::SetSysColors(cElements,lpaElements,lpaRgbValues);
}


//////////////////////////////////////////
/// Window Collection Utility routines ///
//////////////////////////////////////////

// Assimilates the current window into the collective
void CZDDeskFrame::SaveWnd(HWND hWnd)
{
   // Is it a valid window type???
   if(IsCandidate(hWnd))
   {
      // If we shouldn't skip the hide (m_bNoHide is set if we are just building
      // a list of apps for the arrange dialog) and the window is not glued....
      // then hide it
      if(!m_bNoHide && !IsGlued(hWnd))
      {
         // Try to hide the window...if locked up then don't add it to the
         // collective...
         if(!SafeShowWindow(hWnd,SW_HIDE))
            return;
      }

      // Resistence is futile...you've been assimilated
      m_arrHwndDesktop[m_nCurrDesk].Add(hWnd);
   }
}

// Determines if the specified window is a good candidate for assimilation
BOOL CZDDeskFrame::IsCandidate(HWND hWnd)
{
   // One of our excluded windows??
   if(hWnd == m_hWnd || hWnd == m_hWndShell ||   hWnd == m_hWndDesktop ||
                           hWnd == m_hWndPopup)
      return(FALSE);

   // One last chance....is it visible too????
   return(::IsWindowVisible(hWnd));
}

// Searches for the specified window in the normal window list
int CZDDeskFrame::FindNormalWindow(int nDesk, HWND hWnd)
{
   // How many windows we got, anyway??
   int nCount = GetWindowCount(nDesk);

   // Iterate through them all
   for(int i = 0; i < nCount; i++)
   {
      // Aha, we found it...return its index
      if(m_arrHwndDesktop[nDesk][i] == hWnd)
         return(i);
   }

   // Didn't find it, return and error
   return(-1);
}

// Searches for the specifed window in the list of glued windows
int CZDDeskFrame::FindGluedWindow(HWND hWndGlue)
{
   // Get the number of glued windows
   int nCount = GetGluedCount();

   // Iterate through the list
   for(int i = 0; i < nCount; i++)
   {
      // Aha, we got it, tell somebody about it
      if(m_arrHwndGlued[i] == hWndGlue)
         return(i);
   }

   // No dice, try again
   return(-1);
}

// Adds the specified window to the collection objects of each
// desktop
void CZDDeskFrame::AddToAllDesks(HWND hWnd)
{
   int nIndex;

   // Add this window to all desks (even non-configured ones)
   for(int i = 0; i < CZDDeskFrame::MAX_DESKS; i++)
   {
      // Is it already there???
      nIndex = FindNormalWindow(i,hWnd);

      // Not found...add it
      if(nIndex < 0)
         m_arrHwndDesktop[i].Add(hWnd);
   }
}

// Removes the specified window from all desktops except for the one
// specifed by nTab (this is the active tab on the Arrange dialog)
void CZDDeskFrame::RemoveFromAllDesks(HWND hWnd, int nTab)
{
   int nCount;

   // Iterate through all of the desks
   for(int i = 0; i < CZDDeskFrame::MAX_DESKS; i++)
   {
      // On the desk for the current tab??? get out then
      if(i == nTab) continue;
      
      // Get the number of entries in the window collection
      nCount = GetWindowCount(i);

      // Iterate through the window collection until we
      // find the specified window
      for(int j = 0; j < nCount; j++)
      {
         // Found a match??
         if(m_arrHwndDesktop[i][j] == hWnd)
         {
            // Yep, remove it and get out
            m_arrHwndDesktop[i].RemoveAt(j);
            break;
         }
      }
   }
}

// Moves the specifed entry from one desk (nOldDesk) to another desk (nNewDesk)
void CZDDeskFrame::MoveToDesk(int nOldDesk, int nOldEntry, int nNewDesk)
{
   // Get the window handle
   HWND hWnd = m_arrHwndDesktop[nOldDesk][nOldEntry];

   // Remove the window from the current desk...
   m_arrHwndDesktop[nOldDesk].RemoveAt(nOldEntry);

   // And add it to the new desk
   m_arrHwndDesktop[nNewDesk].Add(hWnd);

   // Hide or show the window based on whether we are moving it to or from
   // the currently active desk
   if(nOldDesk == m_nCurrDesk)
      SafeShowWindow(hWnd,SW_HIDE);
   else if(nNewDesk == m_nCurrDesk)
      SafeShowWindow(hWnd,SW_SHOW);
}

// Sets the glue state for a windwo -- if bGlue is TRUE it is glued, if bGlue
// is FALSE it is un-glued
void CZDDeskFrame::SetGlue(HWND hWndGlue, BOOL bGlue, int nTab)
{
   // Try to locate the window
   int nIndex = FindGluedWindow(hWndGlue);

   // Didn't find it and we are glueing....
   if(nIndex < 0 && bGlue)
   {
      // Add it to the glue list and to the window collections
      m_arrHwndGlued.Add(hWndGlue);
      AddToAllDesks(hWndGlue);
   }   // Found and match and not un-glueing
   else if(nIndex >= 0 && !bGlue)  
   {
      // Remove the window from the collection objects
      RemoveFromAllDesks(m_arrHwndGlued[nIndex],nTab);

      // Hide the window on the current desk if it doesn't match
      // the selected tab
      if(nTab != m_nCurrDesk)
         SafeShowWindow(m_arrHwndGlued[nIndex],SW_HIDE);

      // Remove the window from glue collection
      m_arrHwndGlued.RemoveAt(nIndex);
   }
}

// Returns the number of entries in the window collection object for the
// specified desk
int CZDDeskFrame::GetWindowCount(int nDesk)
{
   return(m_arrHwndDesktop[nDesk].GetSize());
}

// Returns the number of entries in the glued window collection object
int CZDDeskFrame::GetGluedCount()
{
   return(m_arrHwndGlued.GetSize());
}

////////////////////////////////////////////////////////////
/// Virtual desktop initialization and shutdown routines ///
////////////////////////////////////////////////////////////

// Initializes the virtual desktop management code: sets up work variables then
// loads default information from the system registry such as desktop colors and
// desktop names
void CZDDeskFrame::InitializeDesks()
{
   // Initialize some work variables
   m_nCurrDesk = 0;
   m_bSwitching = FALSE;
   m_bInMenu = FALSE;
   m_bFrozen = FALSE;
   m_hWndPopup = NULL;
   m_bNoHide = FALSE;
   m_nDeskCount = 5;

   // Construct the system tray interface
   m_pTray = new CZDTray();

   // Get the current background color
   COLORREF clrDesk = ::GetSysColor(COLOR_DESKTOP);

   // Load some default data for colors and desktop names just in case
   // we can't find a registry entry
   CString strTitle;
   for(int i = 0; i < CZDDeskFrame::MAX_DESKS; i++)
   {
      m_arrClrDesk.Add(clrDesk);
      strTitle.Format("Desk %d",i+1);
      m_arrStrDeskName.Add(strTitle);
   }

   // Get the window handles the for taskbar and the desktop (progman)
   m_hWndShell = ::FindWindow("Shell_TrayWnd",NULL);
   m_hWndDesktop = ::FindWindow("Progman",NULL);


   // Initialize the registry stuff
   m_pRegistry = new CZDRegistry();   
   m_pRegistry->Open(HKEY_CURRENT_USER,ZDDESKKEY,KEY_ALL_ACCESS,m_hKey);

   // Get the desktop count
   m_pRegistry->GetValue(m_hKey,"DeskCount",REG_DWORD,&m_nDeskCount,sizeof(m_nDeskCount));

   // Get the desktop names and colors
   char szDeskNameKey[50];
   char szDeskColorKey[50];
   for(i = 0; i < CZDDeskFrame::MAX_DESKS; i++)
   {
      // Key names DeskNamex DeskColorx -- when x is the desktop 1-9
      sprintf(szDeskNameKey,"DeskName%d",i+1);
      sprintf(szDeskColorKey,"DeskColor%d",i+1);

      // Get the name of the desk
      m_pRegistry->GetValue(m_hKey,szDeskNameKey,REG_SZ,
                     m_arrStrDeskName[i].GetBuffer(CZDDeskFrame::MAX_NAME+1),
                     CZDDeskFrame::MAX_NAME);
      m_arrStrDeskName[i].ReleaseBuffer();

      // Get the color of the desk
      m_pRegistry->GetValue(m_hKey,szDeskColorKey,REG_DWORD,&m_arrClrDesk[i],
                                          sizeof(m_arrClrDesk[i]));
   }
}

// Shuts down the virtual desktop management code -- writes the colors and desktop
// names to the registry then makes all hidden apps under ZDDesk's control visible
// again
void CZDDeskFrame::UninitializeDesks()
{
   // Save the desktop count
   m_pRegistry->SetValue(m_hKey,"DeskCount",REG_DWORD,&m_nDeskCount,sizeof(m_nDeskCount));

   char szDeskNameKey[50];
   char szDeskColorKey[50];

   // Save the colors and names
   for(int i = 0; i < CZDDeskFrame::MAX_DESKS; i++)
   {
      sprintf(szDeskNameKey,"DeskName%d",i+1);
      sprintf(szDeskColorKey,"DeskColor%d",i+1);

      // Write the value to the registry
      m_pRegistry->SetValue(m_hKey,szDeskNameKey,REG_SZ,
                        (LPVOID)(LPCTSTR)m_arrStrDeskName[i],
                        strlen(m_arrStrDeskName[i]));

      // If we are still doing color changes then write the current ones out to the
      // registry
      if(!m_bFrozen)
         m_pRegistry->SetValue(m_hKey,szDeskColorKey,REG_DWORD,&m_arrClrDesk[i],
                                          sizeof(m_arrClrDesk[i]));
   }

   // Close the registry and delete the object
   m_pRegistry->Close(m_hKey);
   delete m_pRegistry;

   // Restore all of the virtual desktops by showing all hidden apps on them
   for(m_nCurrDesk = 0; m_nCurrDesk < m_nDeskCount; m_nCurrDesk++)
      LoadDesk(FALSE);

   // Reset the color to the main desktop
   int nColor = COLOR_DESKTOP;
   SafeSetSysColors(1,&nColor,&m_arrClrDesk[0]);

   // Remove the system tray icon and remove the interface object
   m_pTray->Delete(IDI_ZDDESK1);
   delete m_pTray;
}

///////////////////////////////////////////
/// General desktop management routines ///
///////////////////////////////////////////

// Builds the tooltip text name
CString CZDDeskFrame::GetToolTipText()
{
   // Construct tooltip text pre-pended with the application name
   CString strTitle = AfxGetAppName()+CString(" - ");
   strTitle += m_arrStrDeskName[m_nCurrDesk];

   return(strTitle);
}

// The EnumWindows() callback routine:  Gets the window's handle and the 
// class's this pointer
BOOL CALLBACK CZDDeskFrame::EnumCallback(HWND hWnd, LPARAM lParam)
{
   // Get the this pointer
   CZDDeskFrame *pThis = (CZDDeskFrame *) lParam;  

   // This saves window in list (if criteria is met)
   pThis->SaveWnd(hWnd);      

   // Tell Windows to continue the enumeration
   return(TRUE);         
}

// Loads the apps for the specified desktop by performing a ShowWindow
// on each window list entry and waiting a little bit for the window to 
// become visible
void CZDDeskFrame::LoadDesk(BOOL bSetColors)
{
   int nSafety;
   int nCount = GetWindowCount(m_nCurrDesk);
   HWND hWnd;

   // Iterate through the list
   for(int i = 0; i < nCount; i++)
   {
      // Show the window
      hWnd = m_arrHwndDesktop[m_nCurrDesk][i];
      SafeShowWindow(hWnd,SW_SHOW);

      // Wait for the window to draw itself -- give up after a while
      nSafety = 5;
      while(nSafety-- && !::IsWindowVisible(hWnd))
         Sleep(100L);
   }

   // This list is no longer up to date so clean it out now, it will 
   // get refreshed when we switch to another desktop
   m_arrHwndDesktop[m_nCurrDesk].RemoveAll();

   // If we should set colors -- note that setting the color after painting
   // the windows will cause a flash, but we have to do this last in case
   // any windows are "frozen" - i.e the m_bFrozen flag is set because this
   // will cause the WM_SYSCOLORCHANGE message to hang up ZDDesk
   if(bSetColors)
   {
      // Set the background colors
      int nColor = COLOR_DESKTOP;
      SafeSetSysColors(1,&nColor,&m_arrClrDesk[m_nCurrDesk]);
   }
}

// Saves the apps on the current desktop in the window collection by
// calling EnumWindow() -- Note that the EnumCallback() method is 
// reponsible for the "real" work of saving the window info
void CZDDeskFrame::SaveDesk()
{
   // Clear the app list for the current desktop
   m_arrHwndDesktop[m_nCurrDesk].RemoveAll();

   // Enumerate the top level windows
   ::EnumWindows(EnumCallback,(LPARAM)this);

   // Save the system color
   m_arrClrDesk[m_nCurrDesk] = ::GetSysColor(COLOR_DESKTOP);
}

// Handles switching to a new virtual desktop -- called by either a left-mouse
// click message or from a context menu item
void CZDDeskFrame::SwitchDesks(int nNewDesk)
{   
   // Switch desks if necessary
   if(nNewDesk != m_nCurrDesk)
   {
      // Save the current one
      SaveDesk();
      m_nCurrDesk = nNewDesk;

      // Show the new desk's icon and title
      m_pTray->Modify(IDI_ZDDESK1,IDI_ZDDESK1+m_nCurrDesk,GetToolTipText());

      // Load the new desk
      LoadDesk();
   }
}

// Sets the tooltip text for the current active desktop.  This text is based
// on the desktop's name
void CZDDeskFrame::SetCurrentToolTip()
{
   // Tell the tray to change it's title
   m_pTray->Modify(IDI_ZDDESK1,NULL,GetToolTipText());
}

/////////////////////////////////////////////////////////////////////////////
// CZDDeskFrame message handlers

// Handles the WM_CREATE message -- sets up the tray notification window
// and adds the tray icon
int CZDDeskFrame::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
   // Call the base class
   if (CFrameWnd::OnCreate(lpCreateStruct) == -1)
      return -1;

   // Setup the tray information -- notify window, msg and add the try icon
   m_pTray->SetNotifyWindow(GetSafeHwnd(),UM_TRAYNOTIFY);
   m_pTray->Add(IDI_ZDDESK1,GetToolTipText());

   // The following line was added by the Splash Screen component.
   CSplashWnd::ShowSplashScreen(this);

   return 0;
}

// Handles the WM_DESTROY message -- removes the tray notification
void CZDDeskFrame::OnDestroy() 
{
   // Call the base class
   CFrameWnd::OnDestroy();

   // Remove the tray notification
   m_pTray->Delete(IDI_ZDDESK1);
}

// Process the tray notification message -- in our case the UM_TRAYNOTIFY
// message (we specified this in the CZDTray::Add() call in OnCreate() above
LRESULT CZDDeskFrame::OnTrayNotify(WPARAM wParam, LPARAM lParam)
{
   // Determine the tray message
   switch(lParam)
   {
      // Left mouse button release -- send it to the frame's normal message
      // handler
      case WM_LBUTTONUP:
         SendMessage(WM_LBUTTONUP,wParam);
         break;

      // Right mouse button release -- send it to the frame's normal message
      // handler
      case WM_RBUTTONUP:
         SendMessage(WM_RBUTTONUP,wParam);
         break;
   }

   return(0);
}

// Display's the context menu
void CZDDeskFrame::ShowContextMenu()
{
   // Load the context menu
   CMenu menuMain;
   menuMain.LoadMenu(IDR_MENU);
   CMenu *pmenuContext = menuMain.GetSubMenu(0);

   // Make sure our window is activated
   SetForegroundWindow();

   // Add the menu items for the number of desks
   for(int i = 0; i < m_nDeskCount; i++)
      pmenuContext->InsertMenu(i,MF_BYPOSITION,IDM_DESK1+i,
                  m_arrStrDeskName[i]);

   // Get rid of the placeholder
   pmenuContext->DeleteMenu(IDM_DUMMY,MF_BYCOMMAND);


   // Check the current desk...
   pmenuContext->CheckMenuItem(IDM_DESK1+m_nCurrDesk,MF_BYCOMMAND|MF_CHECKED);

   // Display the context menu near the mouse cursor
   POINT pt;
   ::GetCursorPos(&pt);
   pmenuContext->TrackPopupMenu(
               TPM_BOTTOMALIGN|TPM_LEFTBUTTON|TPM_RIGHTBUTTON,
               pt.x,pt.y,this,NULL);

   // We no longer need the menu so destroy it
   menuMain.DestroyMenu();

   // Works around Windows menu handling bug that occurs when no
   // selection was made on the context menu -- the basic bug is that
   // it takes 2 mouse clicks to re-activate the menu after this
   // situation.  Sending a dummy message gets around this
   PostMessage(WM_NULL);
}

// Processes the right mouse button release message -- displays the context menu
void CZDDeskFrame::OnRButtonUp(UINT nFlags, CPoint point) 
{
   if(!m_bInMenu)
   {
      // Display the context menu
      m_bInMenu = TRUE;
      ShowContextMenu();
      m_bInMenu = FALSE;
   }
}

// Handles the "exit" menu option
void CZDDeskFrame::OnExit() 
{
   SendMessage(WM_CLOSE,0,0L);
}

// Processes the left mouse button release message -- switches to a new desktop
void CZDDeskFrame::OnLButtonUp(UINT nFlags, CPoint point) 
{
   // Displaying a menu, or already switching desks???
   if(m_bInMenu || m_bSwitching) 
   {
      // Yep...yell at the user
      MessageBeep(0);
      return;
   }

   // Avoid re-entrancy problems
   m_bSwitching = TRUE;

   // Calculate the new desktop
   int nNewDesk = m_nCurrDesk;
   if(nNewDesk+1 < m_nDeskCount)
      nNewDesk++;
   else nNewDesk = 0;

   // Switch desks
   SwitchDesks(nNewDesk);

   // Clear the re-entrancy flag
   m_bSwitching = FALSE;
}

// Handles the direct desktop selection from the context menu
void CZDDeskFrame::OnSelectDesk(UINT nID)
{
   // Go to the desk selected by the menu item
   SwitchDesks(nID-IDM_DESK1);
}

// Handles the "arrange" menu item
void CZDDeskFrame::OnArrange() 
{
   m_bNoHide = TRUE;
   SaveDesk();
   m_bNoHide = FALSE;

   m_bInMenu = TRUE;

   // Display the arrange dialog
   CZDDeskArrange cArrange(this);
   cArrange.DoModal();

   m_bInMenu = FALSE;
}

// Handles the "properties" menu item
void CZDDeskFrame::OnProperties() 
{
   m_bInMenu = TRUE;

   // Display the properties dialog
   CZDDeskProperties cProperties(this);
   cProperties.DoModal();

   m_bInMenu = FALSE;
}

// Handles the "about" menu item
void CZDDeskFrame::OnAbout() 
{
   m_bInMenu = TRUE;

   // Display the about dialog
   CAbout dlgAbout;
   dlgAbout.DoModal();

   m_bInMenu = FALSE;
}

// Handles the "help" menu item
void CZDDeskFrame::OnHelp() 
{
   // Display the "finder", basically the table of contents
   AfxGetApp()->WinHelp(0,HELP_FINDER);
}


